import inspect
import os
from pathlib import Path

from docutils.core import publish_file


def script_path() -> Path:
    """Return the path of the script being run, whether in a terminal or in a notebook."""
    return Path(os.path.abspath(inspect.stack()[1].filename))


def convert_rst_to_html(input_file: Path, output_file: Path):
    with open(input_file, 'r') as rst_file, open(output_file, 'w') as html_file:
        publish_file(
            source=rst_file,
            destination=html_file,
            writer_name='html'
        )



import inspect
from pathlib import Path
from typing import Optional

import pandas as pd
import plotly
import plotly.offline as pyo
import plotly.graph_objects as go


class MultiPlot:
    def __init__(self, docs_static_dir: Path, tag: Optional[str] = None, super_title: Optional[str] = None,
                 col_wrap: int = 1, save_as_png: bool = False):
        self.tag: Optional[str] = tag
        self.super_title: Optional[str] = super_title
        self.col_wrap: int = col_wrap
        self.save_as_png: bool = save_as_png

        calling_file = inspect.stack()[1].filename

        # Create a subdirectory in docs_static_dir with the stem of the script filename
        self.output_dir: Path = docs_static_dir / Path(calling_file).stem
        self.output_dir.mkdir(parents=True, exist_ok=True)

        # Create the .rst file path
        self.rst_file_path = self.output_dir / f"figures.rst"
        if self.tag:
            self.rst_file_path = self.output_dir / f"figures.{self.tag}.rst"

    def save_plots(self, figs, height: int = 600):
        for i, fig in enumerate(figs):
            suffix: str = f".{i}.html"

            # Save the figure as an HTML file
            file_path = self.output_dir / self.rst_file_path.with_suffix(suffix).name
            pyo.plot(fig, filename=str(file_path), image_height=height, auto_open=False)

            # Open the .rst file, write the iframe, and close the file
            file_mode = 'w' if i == 0 else 'a'
            with open(self.rst_file_path, file_mode) as rst_file:
                if i == 0:
                    if self.super_title is not None:
                        rst_file.write(f".. raw:: html\n\n    <h1>{self.super_title}</h1>\n\n")
                    rst_file.write(".. raw:: html\n\n    <table style='width: 100%;'>\n")
                if i % self.col_wrap == 0:
                    rst_file.write("        <tr>\n")
                rst_file.write(
                    f"            <td><iframe src='../_static/{file_path.parent.name}/{file_path.name}' width='100%' height='{height}px' style='border:1px solid lightgray;'></iframe></td>\n")
                if (i + 1) % self.col_wrap == 0 or i == len(figs) - 1:
                    rst_file.write("        </tr>\n")
                if i == len(figs) - 1:
                    rst_file.write("    </table>\n")

            # Save the figure as a PNG file if save_as_png is True
            if self.save_as_png:
                fig.write_image(file_path.with_suffix('.png'))


def dataframe_to_figure(df: pd.DataFrame) -> go.Figure:
    # Convert the DataFrame to a list of lists
    df_values = df.values.tolist()

    # Create a table
    fig = go.Figure(data=[go.Table(header=dict(values=df.columns.tolist()),
                                   cells=dict(values=df_values))])
    return fig


def dataframe_to_image(df: pd.DataFrame, output_file_name: Path):
    # Create a table
    fig = dataframe_to_figure(df)

    # Save the table as a PNG image
    plotly.write_image(fig, output_file_name)



from abc import ABC

import numpy as np
import pandas as pd
import sklearn
from sklearn.base import RegressorMixin, BaseEstimator, ClassifierMixin, clone
from sklearn.exceptions import NotFittedError
# from sklearn.utils import Tags, TargetTags
from sklearn.utils._estimator_html_repr import _VisualBlock
from sklearn.utils.validation import check_is_fitted
from typing import Union, Optional


# noinspection PyAttributeOutsideInit
class PartitionEstimatorBase(BaseEstimator, ABC):
    def __init__(self, estimator, *, partition_defs: dict[str, str], n_jobs=None, verbose=False):
        super().__init__()

        self.n_jobs: int = n_jobs
        self.verbose: bool = verbose
        self.estimator: BaseEstimator = estimator

        self.estimators: list[tuple[str, object]] = [(k, clone(estimator)) for k in partition_defs.keys()]
        self.partition_defs = partition_defs

        # set post fitting
        self.x_train_: pd.DataFrame
        self.y_train_: pd.DataFrame
        self.domain_indexes_: dict
        self.feature_names_in_: list

    def fit(self, X: pd.DataFrame, y: Union[pd.DataFrame, None] = None):
        # fit the estimators only on the appropriate records
        self.x_train_ = X
        self.y_train_ = y if isinstance(y, pd.DataFrame) else y.to_frame()
        domain_indexes: dict = dict()
        domains: pd.Series = pd.Series(np.nan, index=X.index, name='domains', dtype='object')
        for domain, criteria in self.partition_defs.items():
            mdl = [est[1] for est in self.estimators if est[0] == domain][0]
            xf = X.query(criteria)
            domain_indexes[domain] = xf.index
            if len(domain_indexes[domain]) > 0:
                mdl.fit(xf, y.loc[xf.index])
                domains[xf.index] = domain
        domains = domains.astype('category')
        self.domain_indexes_ = domain_indexes
        self.feature_names_in_ = X.columns.to_list()
        self.domains_ = domains
        self.undomained_ = X.loc[domains.isna() == True]
        return self

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        # predict on records that match the criteria for each model only and merge.
        # nans will be present in the result of the domains do not provide full coverage
        chunks: list = []
        for domain, criteria in self.partition_defs.items():
            mdl = [est[1] for est in self.estimators if est[0] == domain][0]
            xf = X.query(criteria)
            if len(self.domain_indexes_[domain]) > 0:
                check_is_fitted(mdl)
                chunks.append(
                    pd.DataFrame(mdl.predict(xf), index=xf.index, columns=self.y_train_.columns).assign(domain=domain))
            else:
                # this model had no data to fit on, so return nans
                chunks.append(pd.DataFrame(np.nan, index=xf.index, columns=self.y_train_.columns).assign(domain=domain))
        res: pd.DataFrame = pd.merge(left=pd.Series(X.index, index=X.index, name='dummy'),
                                     right=pd.concat(chunks, axis=0),
                                     left_index=True, right_index=True,
                                     how='left', ).drop(columns=['dummy', 'domain'])
        return res

    def score(self, X, y, sample_weight=None):
        # implementation goes here
        pass

    @property
    def n_features_in_(self):
        """Number of features seen during :term:`fit`."""
        # For consistency with other estimators we raise a AttributeError so
        # that hasattr() fails if the estimator isn't fitted.
        try:
            check_is_fitted(self)
        except NotFittedError as nfe:
            raise AttributeError(
                "{} object has no n_features_in_ attribute.".format(
                    self.__class__.__name__
                )
            ) from nfe

        return self.estimators_[0].n_features_in_

    def _more_tags(self):
        return {'estimator_type': 'regressor', 'requires_fit': True}

    if sklearn.__version__ >= '1.6.0':
        def __sklearn_tags__(self):
            from sklearn.utils import Tags, TargetTags
            return Tags(estimator_type=None,
                        target_tags=TargetTags(required=False),
                        transformer_tags=None,
                        regressor_tags=None,
                        classifier_tags=None,
                        )

    def _sk_visual_block_(self):
        names, estimators = zip(*self.estimators)
        criteria = list(self.partition_defs.values())
        return _VisualBlock("parallel", estimators, names=names, name_details=criteria)


class PartitionRegressor(PartitionEstimatorBase, RegressorMixin):
    """Prediction for partitioned subsets of records (estimation domains) defined by filter criteria.

    A PartitionRegressor is designed to allow multiple models to be fitted on a subset of records defined by
     a criteria applied to the input features.  So in essence the criteria definitions define estimation domains
    (a.k.a. partitions) for individual model fitting.
    """

    def __init__(self, estimator, *, partition_defs: dict[str, str], n_jobs=None, verbose=False):
        super().__init__(estimator, partition_defs=partition_defs, n_jobs=n_jobs, verbose=verbose)
        # set scorer for regression
        self.scorer_ = 'r2'
        self._estimator_type = 'regressor'

    def fit(self, X: pd.DataFrame, y: Optional[pd.DataFrame] = None) -> 'PartitionRegressor':
        super().fit(X, y)
        return self

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        return super().predict(X)

    if sklearn.__version__ >= '1.6.0':
        def __sklearn_tags__(self):
            tags = super().__sklearn_tags__()
            tags.estimator_type = 'regressor'
            return tags


class PartitionClassifier(PartitionEstimatorBase, ClassifierMixin):
    """Prediction for partitioned subsets of records (estimation domains) defined by filter criteria.

    A PartitionClassifier is designed to allow multiple models to be fitted on a subset of records defined by
     a criteria applied to the input features.  So in essence the criteria definitions define estimation domains
    (a.k.a. partitions) for individual model fitting.
    """

    def __init__(self, estimator, *, partition_defs: dict[str, str], n_jobs=None, verbose=False):
        super().__init__(estimator, partition_defs=partition_defs, n_jobs=n_jobs, verbose=verbose)
        # set scorer for regression
        self.scorer_ = 'accuracy'
        self._estimator_type = 'classifier'

    def fit(self, X: pd.DataFrame, y: Optional[pd.DataFrame] = None) -> 'PartitionClassifier':
        super().fit(X, y)
        return self

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        return super().predict(X)

    if sklearn.__version__ >= '1.6.0':
        def __sklearn_tags__(self):
            tags = super().__sklearn_tags__()
            tags.estimator_type = 'classifier'
            return tags


import logging
import math
import multiprocessing
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Dict, Any, Union, Optional
from typing import List

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from sklearn.base import BaseEstimator, is_classifier, is_regressor
from sklearn.model_selection import KFold, cross_validate

from elphick.sklearn_viz.model_selection.metrics import classification_metrics, regression_metrics
from elphick.sklearn_viz.model_selection.scorers import classification_scorers, regression_scorers


@dataclass
class CrossValidationResult:
    test_scores: List[float]
    train_scores: List[float]
    fit_times: List[float]
    score_times: List[float]
    estimator: List[Any]
    metrics: Dict[str, List[float]]
    metrics_group: Dict[str, Dict[str, List[float]]]


class CrossValidatorBase(ABC):
    def __init__(self,
                 estimators: Union[BaseEstimator, Dict[str, BaseEstimator]],
                 datasets: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
                 target: str,
                 pre_processor: Optional[Any],
                 cv: Union[int, Any],
                 scorer: Any,
                 metrics: Optional[Dict[str, Any]],
                 group: Any,
                 random_state: int,
                 n_jobs: Union[int, str] = 1,
                 verbosity: int = 1):

        self._logger = logging.getLogger(self.__class__.__name__)

        # If algorithms is not a dictionary, convert it into a dictionary with a default key
        if not isinstance(estimators, dict):
            estimators = {'estimator': estimators}

        # If datasets is not a dictionary, convert it into a dictionary with a default key
        if not isinstance(datasets, dict):
            datasets = {'dataset': datasets}

        # Check if all estimators are classifiers or regressors
        self.is_classifier = all(is_classifier(estimator) for estimator in estimators.values())
        self.is_regressor = all(is_regressor(estimator) for estimator in estimators.values())

        if not self.is_classifier and not self.is_regressor:
            raise ValueError("All estimators must be either classifiers or regressors.")

        if scorer is None:
            scorer = 'accuracy' if self.is_classifier else 'r2'

        if metrics is None:
            default_metrics = {
                'classification': classification_metrics,
                'regression': regression_metrics
            }
            metrics_type = 'classification' if self.is_classifier else 'regression'
            metrics = default_metrics.get(metrics_type)

        self.estimators: Dict[str, BaseEstimator] = estimators
        self.datasets: Dict[str, pd.DataFrame] = datasets
        self.target: str = target
        self.pre_processor: Optional[Any] = pre_processor
        self.cv: Union[int, Any] = cv
        self.scorer: Any = scorer
        self.metrics: Dict[str, Any] = metrics
        self.group: Any = group
        self.random_state: int = random_state
        self.n_jobs: int = n_jobs
        self.verbosity: int = verbosity

        self._logger: logging.Logger = logging.getLogger(self.__class__.__name__)
        self._results: Optional[CrossValidationResult] = None

        self.features_in: List[str] = [col for col in self.datasets[list(self.datasets.keys())[0]] if
                                       col != self.target]

        self._data: Optional[Dict] = None
        self._num_algorithms: int = len(list(self.estimators.keys()))
        self._num_datasets: int = len(list(self.datasets.keys()))

    @property
    def n_cores(self) -> int:
        n_cores = self.n_jobs
        if self.n_jobs < 0:
            n_cores = multiprocessing.cpu_count() + 1 + self.n_jobs
        return n_cores

    @property
    def results(self) -> Optional[Dict[str, Dict[str, CrossValidationResult]]]:
        if self._results is None:
            start_time = datetime.now()  # Record the start time
            if self.verbosity > 0:
                self._logger.info(f"Commencing cross validation...")

            d_results: Dict = {data_key: {algo_key: {} for algo_key in self.estimators.keys()} for data_key in
                               self.datasets.keys()}

            # Run the tasks in a loop
            for data_key, data in self.datasets.items():
                for estimator_key, estimator in self.estimators.items():
                    data_key, estimator_key, res = self.cross_validate_task(data_key, data, estimator_key, estimator)
                    d_results[data_key][estimator_key] = res

            self._results = d_results

            if self.verbosity > 0:
                duration = str(timedelta(seconds=round((datetime.now() - start_time).total_seconds())))
                self._logger.info(f"Cross validation complete in {duration} using {self.n_cores} "
                                  f"worker{'s' if self.n_cores > 1 else ''}")

        return self._results

    def cross_validate_task(self, data_key, data, estimator_key, estimator):
        start_time = datetime.now()  # Record the start time
        if self.verbosity > 1:
            self._logger.info(f"Starting cross-validation for {data_key} with {estimator_key}")

        # Ensure that only the features present in the dataset are used
        features_in_dataset = list(set(self.features_in).intersection(set(data.columns)))
        x: pd.DataFrame = data[features_in_dataset]
        y: pd.DataFrame = data[self.target]
        if self.pre_processor:
            x = self.pre_processor.set_output(transform="pandas").fit_transform(X=x)

        if isinstance(self.cv, int):
            cv = KFold(n_splits=self.cv, random_state=self.random_state, shuffle=True)
        else:
            cv = self.cv
        res = cross_validate(estimator, x, y, cv=cv, scoring=self.scorer, return_train_score=True,
                             return_estimator=True, return_indices=True, error_score=np.nan,
                             n_jobs=self.n_jobs)
        if self.metrics is not None:
            res['metrics'], res['metrics_group'] = self.calculate_metrics(x=x, y=y,
                                                                          estimators=res['estimator'],
                                                                          indices=res['indices'],
                                                                          group=self.group)
        if self.verbosity > 1:
            duration = str(timedelta(seconds=round((datetime.now() - start_time).total_seconds())))
            res_mean = res[f"test_score"].mean()
            res_std = res[f"test_score"].std()
            # Format the duration
            self._logger.info(f"Finished cross-validation for {data_key} with {estimator_key}."
                              f" Mean = {res_mean}, SD = {res_std}, Duration: {duration}")

        # Convert the results to a CrossValidationResult instance and return it
        return data_key, estimator_key, CrossValidationResult(
            test_scores=res['test_score'],
            train_scores=res['train_score'],
            fit_times=res['fit_time'],
            score_times=res['score_time'],
            estimator=res['estimator'],
            metrics=res['metrics'],
            metrics_group=res['metrics_group']
        )

    @abstractmethod
    def get_cv_scores(self):
        chunks: List = []
        for data_key, data in self.datasets.items():
            for estimator_key, estimator in self.estimators.items():
                chunks.append(
                    pd.Series(self.results[data_key][estimator_key].test_scores, name=(data_key, estimator_key)))
        return pd.concat(chunks, axis=1)

    @abstractmethod
    def get_cv_metrics(self, metrics, by_group: bool = False) -> pd.DataFrame:
        chunks: List = []
        metric_key = "metrics_group" if by_group else "metrics"
        for data_key, data in self.datasets.items():
            for estimator_key, estimator in self.estimators.items():
                for metric in metrics:
                    # Access the metrics or metrics_group property of the CrossValidationResult instance
                    metric_data = getattr(self.results[data_key][estimator_key], metric_key)[metric]
                    chunks.append(pd.DataFrame(metric_data).assign(
                        **dict(data_key=data_key, algo_key=estimator_key, metric=metric)))
        res: pd.DataFrame = pd.concat(chunks, axis=0).set_index(['data_key', 'algo_key'], append=True).rename(
            columns={0: 'value'})
        res.index.names = ['fold', 'data_key', 'algo_key']
        return res

    @abstractmethod
    def calculate_metrics(self, x, y, estimators, indices, group):
        metric_results: Dict = {}
        metric_results_group: Dict = {}

        for k, fn_metric in self.metrics.items():
            metric_values: List = []
            metric_groups: Dict = {}
            for estimator, test_indexes in zip(estimators, indices['test']):
                y_true = y[y.index[test_indexes]]
                y_est = estimator.predict(x.loc[x.index[test_indexes], :])
                if isinstance(y_est, pd.DataFrame) and y_est.shape[1] == 1:
                    y_est = y_est.iloc[:, 0]
                metric_values.append(fn_metric(y_true, y_est))
                if group is not None:
                    # calculate the metric by each group in the group series.
                    y_est = pd.merge(left=pd.Series(y_est, name='y_est', index=x.index[test_indexes]),
                                     right=group, left_index=True, right_index=True)
                    y_est_grouped = y_est.groupby([group.name], observed=False)
                    grouped_results = [y_est_grouped.get_group((x,)) for x in y_est_grouped.groups]
                    for grp_res in grouped_results:
                        group_value = str(grp_res[group.name].iloc[0])
                        group_metric_results = fn_metric(y_true[grp_res.index], grp_res['y_est'].values)
                        if group_value not in metric_groups.keys():
                            metric_groups[group_value] = [group_metric_results]
                        else:
                            metric_groups[group_value].append(group_metric_results)
            metric_results[k] = metric_values
            if group is not None:
                metric_results_group[k] = metric_groups

        return metric_results, metric_results_group

    def get_model_by_group_data(self, estimator, dataset) -> tuple[pd.DataFrame, pd.DataFrame]:
        results: dict = {}
        by_group_score_chunks: list = []
        by_group_metric_chunks: list = []
        for grp in self.group.unique():
            grp_index: pd.Index = self.group.loc[self.group == grp].index
            x: pd.DataFrame = self.datasets[dataset][self.features_in].loc[grp_index]
            y: pd.DataFrame = self.datasets[dataset][self.target].loc[grp_index]
            if self.pre_processor:
                x = self.pre_processor.set_output(transform="pandas").fit_transform(X=x)
            cv = self.cv
            if isinstance(self.cv, int):
                cv = KFold(n_splits=self.cv, random_state=self.random_state, shuffle=True)

            res = cross_validate(self.estimators[estimator], x, y, cv=cv, scoring=self.scorer, return_estimator=True,
                                 return_indices=True, n_jobs=self.n_jobs)
            by_group_score_chunks.append(pd.Series(res['test_score'], name=grp))
            if self.metrics is not None:
                res['metrics'], _ = self.calculate_metrics(x=x, y=y,
                                                           estimators=res['estimator'],
                                                           indices=res['indices'],
                                                           group=None)
            results[grp] = res
            by_group_metric_chunks.append(
                pd.DataFrame(res['metrics']).assign(group=grp).rename_axis('fold', axis='index'))
        by_group_metrics: pd.DataFrame = pd.concat(by_group_metric_chunks)
        by_group_scores = pd.concat(by_group_score_chunks, axis=1)

        return by_group_metrics, by_group_scores

    @abstractmethod
    def plot(self, metrics=None, show_group=False, title=None, col_wrap=None):
        pass



from typing import Optional, List

import pandas as pd

import plotly.graph_objects as go


def plot_parallel_coordinates(data: pd.DataFrame,
                              color: Optional[str] = None,
                              vars_include: Optional[List[str]] = None,
                              vars_exclude: Optional[List[str]] = None,
                              title: Optional[str] = None,
                              ) -> go.Figure:
    """Create an interactive parallel plot

    Useful to explore multidimensional data like mass-composition data

    Args:
        data: The DataFrame to plot
        color: The variable that sets the color, typically the target variable
        vars_include: Optional List of variables to include in the plot
        vars_exclude: Optional List of variables to exclude in the plot
        title: Optional plot title

    Returns:

    """
    df: pd.DataFrame = data.copy()
    if vars_include is not None:
        missing_vars = set(vars_include).difference(set(df.columns))
        if len(missing_vars) > 0:
            raise KeyError(f'vars_include provided contains variable not found in the data: {missing_vars}')
        df = df[vars_include]
    if vars_exclude:
        df = df[[col for col in df.columns if col not in vars_exclude]]

    # Kudos: https://stackoverflow.com/questions/72125802/parallel-coordinate-plot-in-plotly-with-continuous-
    # and-categorical-data

    categorical_columns = df.select_dtypes(include=['category', 'object'])
    col_list = []

    for col in df.columns:
        if col in categorical_columns:  # categorical columns
            cat_map: dict = dict(enumerate(df[col].cat.categories))
            df[col] = df[col].cat.codes
            col_dict = dict(
                label=col,
                tickvals=list(cat_map.keys()),
                ticktext=list(cat_map.values()),
                values=df[col],
            )
        else:  # continuous columns
            col_dict = dict(
                range=(df[col].min(), df[col].max()),
                label=col,
                values=df[col],
            )
        col_list.append(col_dict)

    if color is None:
        fig = go.Figure(data=go.Parcoords(dimensions=col_list))
    else:
        fig = go.Figure(data=go.Parcoords(dimensions=col_list, line=dict(color=df[color])))

    fig.update_layout(title=title, height=700)

    return fig



from typing import Optional

import pandas as pd
import plotly.graph_objects as go
import plotly.express as px


def plot_scatter_matrix(x: pd.DataFrame, color: Optional[str] = None, title: Optional[str] = None) -> go.Figure:
    """Plot a scatter matrix

    Args:
        x: X values to transform and plot.
        color: optional target vector
        title: Optional plot title

    Returns:

    """

    fig = px.scatter_matrix(data_frame=x, dimensions=list(x.columns), color=color)
    fig.update_traces(diagonal_visible=False)
    title = 'Scatter Matrix' if title is None else title
    fig.update_layout(title=title)

    return fig
